# frozen_string_literal: true

require_relative "../lib/wayfarer"

require "benchmark"
require "parallel" unless RUBY_PLATFORM == "java"
require "faker"

include Wayfarer::Frontiers

def print_separator
  puts "-" * `tput cols`.to_i
end

# URI count is kept low on purpose because this file is run on Travis CI.
# Bump up the numbers when running locally.
URI_COUNT = 10_000
SAMPLE_COUNT = URI_COUNT * 0.01
CYCLE_COUNT = 20

FRONTIERS = RUBY_PLATFORM == "java" ? {
  MemoryFrontier => "memory",
  RedisFrontier => "redis"
} : {
  MemoryBloomfilter => "memory_bloom",
  MemoryFrontier => "memory",
  MemoryTrieFrontier => "memory_trie",
  RedisBloomfilter => "redis_bloom",
  RedisFrontier => "redis"
}

label_width = FRONTIERS.reduce(0) { |acc, (_, label)|
  acc > (l = label.length) ? acc : l
}

puts "Generating #{URI_COUNT} URIs..."

uris = if RUBY_PLATFORM == "java"
         Array.new(URI_COUNT) { Faker::Internet.url }
       else
         Parallel.map(URI_COUNT.times) { Faker::Internet.url }
       end
samples = uris.sample(SAMPLE_COUNT)

puts "Done generating URIs"

print_separator
puts "Cache insertion"

Benchmark.bm do |bm|
  FRONTIERS.each do |(klass, label)|
    frontier = klass.new(Wayfarer::Configuration.new)
    frontier.cache(*uris)

    bm.report(label.ljust(label_width)) do
      frontier.cache(*samples)
    end

    frontier.free
  end
end

print_separator
puts "Cache detection"

Benchmark.bm do |bm|
  FRONTIERS.each do |(klass, label)|
    frontier = klass.new(Wayfarer::Configuration.new)
    frontier.cache(*uris)

    bm.report(label.ljust(label_width)) do
      samples.each { |sample| frontier.cached?(sample) }
    end

    frontier.free
  end
end

print_separator
puts "Cycling"

slices = samples.each_slice(CYCLE_COUNT)

Benchmark.bm do |bm|
  FRONTIERS.each do |(klass, label)|
    frontier = klass.new(Wayfarer::Configuration.new)

    bm.report(label.ljust(label_width)) do
      slices.each do |samples|
        frontier.stage(*samples)
        frontier.cycle
      end
    end

    frontier.free
  end
end

print_separator
puts "Cache detection false positives/negatives"

trials = samples.count / 2

positives = samples.sample(trials)
negatives = if RUBY_PLATFORM == "java"
              Array.new(trials) { Faker::Internet.url }
            else
              Parallel.map(trials.times) { Faker::Internet.url }
            end

FRONTIERS.each do |(klass, label)|
  frontier = klass.new(Wayfarer::Configuration.new)

  false_positives = false_negatives = 0

  positives.each do |positive| frontier.cache(positive) end

  positives.each do |positive|
    false_positives += 1 unless frontier.cached?(positive)
  end

  negatives.each do |negative|
    false_positives += 1 if frontier.cached?(negative)
  end

  fp_percentage = if false_positives.zero?
                    0
                  else
                    (trials.to_f / false_positives).round(3)
                  end

  fn_percentage = if false_negatives.zero?
                    0
                  else
                    (trials.to_f / false_negatives).round(3)
                  end

  puts "* #{label.ljust(label_width)}"
  puts "  * False positives: #{false_positives} (#{fp_percentage}%)"
  puts "  * False negatives: #{false_negatives} (#{fn_percentage}%)"

  frontier.free
end
