# frozen_string_literal: true

module Wayfarer
  class CrawlObserver
    module Events
      FirstCycle = Struct.new(:frontier)
      NewCycle = Struct.new(:current_uris_count)
      DispatchedURI = Struct.new(:action, :uri)
      CycleFinished = Class.new
      Peeking = Struct.new(:uri)
      AboutToCycle = Struct.new(:staged_uris_count)
      MismatchedURI = Struct.new(:uri)
      HaltInitiated = Struct.new(:action, :uri)
      StagingURIs = Struct.new(:staged_uris_count)
      UnhandledError = Struct.new(:exception)
    end

    module ObservableShortcuts
      def notify_observers!(*argv)
        changed
        notify_observers(*argv)
      end
    end

    extend Forwardable

    attr_reader :logger

    def initialize(*observables, logger)
      @logger = logger
      observables.each { |obsv| obsv.add_observer(self) }
    end

    def update(event)
      case event
      when Events::FirstCycle     then first_cycle(event)
      when Events::NewCycle       then new_cycle(event)
      when Events::DispatchedURI  then dispatched_uri(event)
      when Events::CycleFinished  then cycle_finished
      when Events::Peeking        then peeking(event)
      when Events::AboutToCycle   then about_to_cycle(event)
      when Events::MismatchedURI  then mismatched_uri(event)
      when Events::HaltInitiated  then halt_initiated(event)
      when Events::StagingURIs    then staging_uris(event)
      when Events::UnhandledError then unhandled_error(event)
      end
    end

    private

    def first_cycle(event)
      logger.info("First cycle")
      logger.info("Frontier: #{event.frontier}")
    end

    def new_cycle(event)
      logger.info("Current cycle contains #{event.current_uris_count} URI(s)")
    end

    def dispatched_uri(event)
      logger.info("Dispatched to \##{event.action}: #{event.uri}")
    end

    def cycle_finished
      logger.info("No URIs left in current cycle")
    end

    def peeking(event)
      logger.info("Peeking into: #{event.uri}")
    end

    def about_to_cycle(event)
      logger.info("About to cycle. #{event.staged_uris_count} staged URI(s)")
    end

    def mismatched_uri(event)
      logger.debug("No matching route for: #{event.uri}")
    end

    def halt_initiated(event)
      logger.info("Halt initiated from \##{event.action} at: #{event.uri}")
    end

    def staging_uris(event)
      logger.info("Staging #{event.staged_uris_count} URI(s)")
    end

    def unhandled_error(event)
      level = config.reraise_exceptions ? :fatal : :error

      if config.print_stacktraces
        logger.public_send level, <<~LOGGER
          Unhandled exception in an action: #{event.exception.class.inspect}
          #{event.exception.backtrace.map(&:to_s).join("\n* ")}
        LOGGER
      else
        logger.public_send level, <<~LOGGER
          Unhandled exception in an action: #{event.exception.class.inspect}
        LOGGER
      end
    end
  end
end
