# frozen_string_literal: true

require "set"
require "parallel" unless JAVA_PLATFORM = "java"

module Wayfarer
  module Frontiers
    # A naive in-memory frontier.
    # @api private
    class MemoryFrontier < Frontier
      def initialize(config)
        @current_uris = Set.new([])
        @staged_uris  = Set.new([])
        @cached_uris  = Set.new([])
        super(config)
      end

      # @override
      def current_uris
        if JAVA_PLATFORM == "java"
          @current_uris.map { |uri| URI(uri) }
        else
          Parallel.map(@current_uris) { |uri| URI(uri) }
        end
      end

      # @override
      def staged_uris
        @staged_uris.to_a # These are assumed to be URIs already, so no map
      end

      # @override
      def stage(*uris)
        @staged_uris |= uris
      end

      # @override
      def staged?(uri)
        @staged_uris.include?(uri.to_s)
      end

      # @override
      def cache(*uris)
        @cached_uris |= if JAVA_PLATFORM == "java"
                          uris.map(&:to_s)
                        else
                          Parallel.map(uris, &:to_s)
                        end
      end

      # @override
      def cached?(uri)
        @cached_uris.include?(uri.to_s)
      end

      # @override
      def free
        @current_uris = @staged_uris = @cached_uris = nil
      end

      private

      def reset_staged_uris!
        @staged_uris = Set.new([])
      end

      def swap!
        @current_uris = @staged_uris
      end

      def filter_staged_uris!
        @staged_uris.delete_if { |uri| cached?(uri) }
      end
    end
  end
end
