# frozen_string_literal: true

require "selenium-webdriver"
require "selenium/emulated_features"
require "capybara"

module Wayfarer
  module HTTPAdapters
    # An adapter for Selenium WebDrivers
    # @api private
    class SeleniumAdapter
      # @!attribute [r] driver
      # @return [URI] the Selenium WebDriver.
      attr_reader :driver

      def initialize(config = Wayfarer.config)
        @config = config
      end

      # Fetches a page.
      # @return [Page]
      def fetch(uri)
        driver.navigate.to(uri)

        Page.new(
          uri: @driver.current_url,
          status_code: @driver.response_code,
          body: @driver.page_source,
          headers: @driver.response_headers
        )
      end

      # Closes the driver.
      def reload!
        @driver&.close
        @driver = nil
      end

      # Quits the browser.
      def free
        @driver&.quit
        @driver = nil
      end

      # The WebDriver.
      def driver
        @driver ||= instantiate_driver
      end

      # A Capybara driver that wraps the {#driver}.
      # @see https://github.com/teamcapybara/capybara Capybara
      def browser
        @browser ||= instantiate_capybara_driver
      end

      private

      def instantiate_driver
        driver = Selenium::WebDriver.for(*@config.selenium_argv)
        driver.manage.window.size = Selenium::WebDriver::Dimension.new(
          *@config.window_size
        )
        driver
      end

      def instantiate_capybara_driver
        Capybara.run_server = false
        Capybara.current_driver = :selenium

        capybara_driver = Capybara::Selenium::Driver.new(nil)
        capybara_driver.instance_variable_set(:@browser, driver)

        session = Capybara::Session.new(:selenium, nil)
        session.instance_variable_set(:@driver, capybara_driver)

        session
      end
    end
  end
end
