# frozen_string_literal: true

require "pp"
require "concurrent"
require "observer"

module Wayfarer
  # Runs jobs.
  class Processor
    extend Forwardable

    include Observable
    include CrawlObserver::Events
    include CrawlObserver::ObservableShortcuts

    attr_reader :job

    delegate config: :job
    delegate logger: :config

    def initialize(job, frontier, dispatcher)
      @job = job
      @frontier = frontier
      @dispatcher = dispatcher
      @halted = Concurrent::AtomicBoolean.new(false)
    end

    # Whether processing is done.
    # @return [true, false]
    def halted?
      @halted.value
    end

    # Sets a halt flag.
    def halt!
      @halted.make_true
    end

    # Runs the job.
    # @param [*Array<URI>, *Array<String>] uris
    def run(*_uris)
      notify_observers!(FirstCycle.new(@frontier))

      while @halted.false? && @frontier.cycle
        current_uris = @frontier.current_uris
        queue = current_uris.inject(Queue.new, :push)

        notify_observers!(NewCycle.new(current_uris.count))

        @threads = Array.new(config.connection_count) do
          Thread.new do
            begin
              loop do
                uri = queue.pop(true)
                break if uri.nil? || @halted.true?
                handle_dispatch_result(@dispatcher.dispatch(@job, uri))
              end
            rescue ThreadError
              notify_observers!(CycleFinished.new)
            end
          end
        end

        @threads.each(&:join)

        notify_observers!(AboutToCycle.new(@frontier.staged_uris.count))
      end
    ensure
      halt!
      @frontier.free
      @dispatcher.adapter_pool.free
    end

    private

    def handle_dispatch_result(result)
      case result
      when Dispatcher::Mismatch then handle_mismatch(result)
      when Dispatcher::Halt     then handle_halt(result)
      when Dispatcher::Stage    then handle_stage(result)
      when Dispatcher::Error    then handle_error(result)
      end
    end

    def handle_mismatch(mismatch)
      notify_observers!(MismatchedURI.new(mismatch.uri))
    end

    def handle_halt(halt)
      notify_observers!(HaltInitiated.new(halt.action, halt.uri))
      halt!
    end

    def handle_stage(stage)
      notify_observers!(StagingURIs.new(stage.uris.count))
      @frontier.stage(*stage.uris) unless halted?
    end

    def handle_error(error)
      notify_observers!(UnhandledError.new(error.exception))
    end
  end
end
