# frozen_string_literal: true

require "forwardable"

module Wayfarer
  module Routing
    # A {Router} maps URIs onto a {Job}'s instance methods.
    class Router
      extend Forwardable

      # @!attribute [r] rule
      # @return [Rule]
      attr_reader :rule

      # @!attribute [r] blacklist
      # @return [Rule]
      attr_reader :blacklist

      def initialize
        @rule = Rule.new
        @blacklist = Rule.new
      end

      delegate %i[
        uri
        host
        path
        query
      ] => :rule

      # Returns the associated instance method (action) of the first rule that
      # matches a URI and the collected parameter hash from the rule chain.
      # @return [[Boolean, Symbol, Hash]] if a matching rule exists.
      # @return [false] if no matching rule exists or the URI is forbidden.
      def route(uri)
        return false if forbids?(uri)

        # TODO: Use structs instead
        is_matching, params, action = @rule.invoke(uri)
        return action, params if is_matching && params

        false
      end

      # Whether a route matches the URI.
      # TODO: Test
      def routes?(uri)
        !!route(uri)
      end

      # Adds a {Rule} to the blacklist.
      def forbid(opts = {}, &proc)
        @blacklist.build_child_rule_chain_from_options(opts)
        @blacklist.instance_eval(&proc) if block_given?
        @blacklist
      end

      # Whether the URI is matched by the blacklist rule.
      # @see #forbid
      def forbids?(uri)
        @blacklist.matches?(uri)
      end

      # Whether the URI is allowed.
      # @see #forbid
      def allows?(uri)
        !forbids?(uri)
      end
    end
  end
end
